from ftw.builder.testing import BUILDER_LAYER
from ftw.builder.testing import functional_session_factory
from ftw.builder.testing import set_builder_session_factory
from plone.app.testing import FunctionalTesting
from plone.app.testing import IntegrationTesting
from plone.app.testing import PLONE_FIXTURE
from plone.app.testing import PloneSandboxLayer
from plone.app.testing import applyProfile
from plone.dexterity.fti import DexterityFTI
from plone.testing import z2
from zope.configuration import xmlconfig
import ftw.solr.tests.builders


class SolrLayer(PloneSandboxLayer):

    defaultBases = (PLONE_FIXTURE, BUILDER_LAYER)

    def setUpZope(self, app, configurationContext):
        import z3c.autoinclude
        xmlconfig.file('meta.zcml', z3c.autoinclude,
                       context=configurationContext)
        xmlconfig.string(
            '<configure xmlns="http://namespaces.zope.org/zope">'
            '  <includePlugins package="plone" />'
            '  <includePluginsOverrides package="plone" />'
            '  <include package="collective.indexing" />'
            '</configure>',
            context=configurationContext)

        z2.installProduct(app, 'collective.indexing')

    def setUpPloneSite(self, portal):
        applyProfile(portal, 'plone.app.dexterity:default')
        applyProfile(portal, 'ftw.solr:default')

        fti = DexterityFTI('DexterityFolder',
                           klass="plone.dexterity.content.Container",
                           global_allow=True)
        portal.portal_types._setObject('DexterityFolder', fti)
        fti.lookupSchema()


SOLR_FIXTURE = SolrLayer()
SOLR_INTEGRATION_TESTING = IntegrationTesting(
    bases=(SOLR_FIXTURE, ), name="ftw.solr:integration")
SOLR_FUNCTIONAL_TESTING = FunctionalTesting(
    bases=(SOLR_FIXTURE,
           set_builder_session_factory(functional_session_factory)),
    name="ftw.solr:functional")



from ftw.solr import _
from plone.autoform import directives
from plone.autoform.interfaces import IFormFieldProvider
from plone.indexer import indexer
from plone.supermodel import model
from plone.supermodel.model import Schema
from z3c.form.interfaces import IAddForm
from z3c.form.interfaces import IEditForm
from zope.i18nmessageid import MessageFactory
from zope.interface import alsoProvides
from zope.schema import Bool
from zope.schema import Text


_plone = MessageFactory('plone')


class IShowInSearch(Schema):
    """Behavior adding a "Show in search" checkbox in order to make
    it possible to exclude certain content from the search results.
    """

    model.fieldset(
        'settings',
        label=_plone(u'Settings'),
        fields=['showinsearch'])

    directives.write_permission(showinsearch='ftw.solr.EditSearchSettings')
    directives.omitted('showinsearch')
    directives.no_omit(IEditForm, 'showinsearch')
    directives.no_omit(IAddForm, 'showinsearch')
    showinsearch = Bool(
        title=_(u'label_showinsearch', default=u'Show in search'),
        description=u'',
        default=True)


alsoProvides(IShowInSearch, IFormFieldProvider)


class ISearchwords(Schema):
    """Adds a "Searchwords" field for adding additional search words
    the the search index for improving search results.
    """

    model.fieldset(
        'settings',
        label=_plone(u'Settings'),
        fields=['searchwords'])

    directives.write_permission(searchwords='ftw.solr.EditSearchSettings')
    directives.omitted('searchwords')
    directives.no_omit(IEditForm, 'searchwords')
    directives.no_omit(IAddForm, 'searchwords')
    searchwords = Text(
        title=_(u'label_searchwords', default=u'Search words'),
        description=_(u'help_searchwords',
                      default=u'Specify words for which this item will show up '
                      u'as the first search result. Multiple words can be '
                      u'specified on new lines.'),
        required=False)

alsoProvides(ISearchwords, IFormFieldProvider)


@indexer(ISearchwords)
def searchwords_indexer(obj):
    words = getattr(ISearchwords(obj), 'searchwords', '')
    if not isinstance(words, unicode):
        words = words.decode('utf-8')

    return filter(None, [word.strip('\r ') for word in words.split('\n')])



from ftw.solr.tests import FunctionalTestCase
from ftw.testbrowser import browsing
from ftw.testbrowser.pages import factoriesmenu
from plone.indexer.interfaces import IIndexer
from Products.CMFCore.utils import getToolByName
from zope.component import getMultiAdapter
from zope.component.hooks import getSite
import transaction


def index_value_for(obj, index_name):
    catalog = getToolByName(obj, 'portal_catalog')
    indexer = getMultiAdapter((obj, catalog), IIndexer, name=index_name)
    return indexer()


def enable_behavior(behavior, type='DexterityFolder'):
    portal_types = getToolByName(getSite(), 'portal_types')
    portal_types['DexterityFolder'].behaviors.append(behavior)
    transaction.commit()


def disable_behavior(behavior, type='DexterityFolder'):
    portal_types = getToolByName(getSite(), 'portal_types')
    portal_types['DexterityFolder'].behaviors.remove(behavior)
    transaction.commit()



class TestShowInSearch(FunctionalTestCase):

    def setUp(self):
        super(TestShowInSearch, self).setUp()
        enable_behavior('ftw.solr.behaviors.IShowInSearch')

    def tearDown(self):
        super(TestShowInSearch, self).tearDown()
        disable_behavior('ftw.solr.behaviors.IShowInSearch')

    @browsing
    def test_True_by_default(self, browser):
        self.grant('Manager')
        browser.login().open()
        factoriesmenu.add('DexterityFolder')
        browser.click_on('Save')
        self.assertTrue(index_value_for(browser.context, 'showinsearch'),
                        'Expected showinsearch to be True by default.')

    @browsing
    def test_setting_to_False(self, browser):
        self.grant('Manager')
        browser.login().open()
        factoriesmenu.add('DexterityFolder')
        browser.fill({'Show in search': False}).save()
        self.assertFalse(index_value_for(browser.context, 'showinsearch'),
                         'Expected showinsearch to be set to False.')


class TestSearchwords(FunctionalTestCase):

    def setUp(self):
        super(TestSearchwords, self).setUp()
        enable_behavior('ftw.solr.behaviors.ISearchwords')

    def tearDown(self):
        super(TestSearchwords, self).tearDown()
        disable_behavior('ftw.solr.behaviors.ISearchwords')

    @browsing
    def test_searchwords_are_provided_by_indexer(self, browser):
        self.grant('Manager')
        browser.login().open()
        factoriesmenu.add('DexterityFolder')
        browser.fill({'Search words': 'Foo Bar\nBaz'}).save()
        browser.open(browser.context, view='edit')
        self.assertEquals([u'Foo Bar', u'Baz'],
                          index_value_for(browser.context, 'searchwords'))



from collective.solr.exceptions import SolrConnectionException
from collective.solr.interfaces import ISolrConnectionManager
from Products.Five.browser import BrowserView
from zope.component import getUtility
import json
import urllib


class SuggestView(BrowserView):
    """Get autocomplete suggestions using solr's suggester component."""

    def __call__(self):
        suggestions = []
        term = self.request.form.get('term', '')
        if not term:
            return json.dumps(suggestions)

        manager = getUtility(ISolrConnectionManager)
        connection = manager.getConnection()

        if connection is None:
            return json.dumps(suggestions)

        params = {}
        params['q'] = term
        params['wt'] = 'json'

        params = urllib.urlencode(params, doseq=True)

        try:
            response = connection.doPost(
                connection.solrBase + '/suggest?' + params, '', {})

        except SolrConnectionException, e:
            return json.dumps([])

        results = json.loads(response.read())
        spellcheck = results.get('spellcheck', None)
        if not spellcheck:
            return json.dumps(suggestions)
        spellcheck_suggestions = spellcheck.get('suggestions', None)
        if not spellcheck_suggestions:
            return json.dumps(suggestions)

        for suggestion in spellcheck_suggestions[1]['suggestion']:
            suggestions.append(dict(label=suggestion, value=suggestion))

        return json.dumps(suggestions)



from Acquisition import aq_inner
from plone.app.portlets.portlets import base
from plone.portlets.interfaces import IPortletDataProvider
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from zope import schema
from zope.component import getMultiAdapter
from zope.formlib import form
from zope.interface import implements
from ftw.solr import _


class IWordCloudPortlet(IPortletDataProvider):
    """A WordCloud portlet.

    It inherits from IPortletDataProvider because for this portlet, the
    data that is being rendered and the portlet assignment itself are the
    same.
    """

    portlet_title = schema.TextLine(
        title=_(u'label_portlet_title', default=u"Title"),
        description=_(u'help_portlet_title', default=u"Portlet title"),
        required=False,
        default=u"Word Cloud")

    num_words = schema.Int(
        title=_(u'label_num_words', default=u"Number of words"),
        description=_(u'help_num_words',
                      default=u"Maximum number of words to be displayed"),
        required=True,
        default=10)

    num_sizes = schema.Int(
        title=_(u'label_num_sizes', u"Number of sizes"),
        description=_(u'help_num_sizes',
                      default=u"Number of different sizes to represent term "
                               "weight"),
        required=True,
        default=5)

    scale_factor = schema.Float(
        title=_(u'label_scale_factor', default=u"Scale factor"),
        description=_(u'help_scale_factor',
                      default=u"Constant scaling factor to scale font size by"),
        required=True,
        default=2.0)


class Assignment(base.Assignment):
    """Portlet assignment for the Word Cloud portlet.

    This is what is actually managed through the portlets UI and associated
    with columns.
    """

    implements(IWordCloudPortlet)

    def __init__(self, num_words, num_sizes, portlet_title, scale_factor):
        self.portlet_title = portlet_title
        self.num_words = num_words
        self.num_sizes = num_sizes
        self.scale_factor = scale_factor

    @property
    def title(self):
        """This property is used to give the title of the portlet in the
        'manage portlets' screen.

        Note: This is different from the `portlet_title` which, if set,
        will be displayed in the portlet header.
        """
        return "Word Cloud"


class Renderer(base.Renderer):
    """Portlet renderer for the Word Cloud portlet.

    This is registered in configure.zcml. The referenced page template only
    contains the basic portlet structure, the actual word cloud is rendered
    in the render_cloud() method by looking up the `wordcloud` browser view.
    """

    render = ViewPageTemplateFile('wordcloud_portlet.pt')

    def header(self):
        portlet_title = getattr(self.data, 'portlet_title', None)
        return portlet_title

    @property
    def available(self):
        return True

    def render_cloud(self):
        context = aq_inner(self.context)
        word_cloud_view = getMultiAdapter((context, self.request),
                                          name='wordcloud')

        num_words = getattr(self.data, 'num_words', None)
        num_sizes = getattr(self.data, 'num_sizes', None)
        scale_factor = getattr(self.data, 'scale_factor', None)

        kwargs = {'num_words': num_words,
                  'num_sizes': num_sizes,
                  'scale_factor': scale_factor}

        return word_cloud_view(**kwargs)


class AddForm(base.AddForm):
    """Portlet add form for the Word Cloud portlet.

    This is registered in configure.zcml. The form_fields variable tells
    zope.formlib which fields to display. The create() method actually
    constructs the assignment that is being added.
    """
    form_fields = form.Fields(IWordCloudPortlet)

    def create(self, data):
        return Assignment(**data)


class EditForm(base.EditForm):
    """Portlet edit form.

    This is registered with configure.zcml. The form_fields variable tells
    zope.formlib which fields to display.
    """
    form_fields = form.Fields(IWordCloudPortlet)



# -*- coding: utf-8 -*-
from collective.solr.exceptions import SolrConnectionException
from ftw.solr.interfaces import ISearchSettings
from logging import getLogger
from plone.app.contentlisting.interfaces import IContentListing
from plone.app.search import browser
from plone.registry.interfaces import IRegistry
from Products.CMFCore.utils import getToolByName
from Products.CMFPlone.PloneBatch import Batch
from Products.CMFPlone.utils import safe_hasattr
from Products.Five.browser.pagetemplatefile import ViewPageTemplateFile
from Products.ZCTextIndex.ParseTree import ParseError
from zope.component import getMultiAdapter
from zope.component import getUtility
from zope.deferredimport import deprecated
import urllib
import urlparse


logger = getLogger('ftw.solr')


deprecated(
    "This class is moved to another place. "
    "Please use ftw.solr.viewlets.searchbox.SearchBoxViewlet instead",
    SearchBoxViewlet='ftw.solr.viewlets.searchbox:SearchBoxViewlet')


class SearchView(browser.Search):

    template = ViewPageTemplateFile('search.pt')
    results_template = ViewPageTemplateFile('results.pt')

    def __call__(self):
        if 'ajax' in self.request.form:
            del self.request.form['ajax']
            # Disable theming for ajax requests
            self.request.response.setHeader('X-Theme-Disabled', 'True')
            return self.results_template()
        return self.template()

    def render_results(self):
        return self.results_template()

    def filter_query(self, query):
        registry = getUtility(IRegistry)
        settings = registry.forInterface(ISearchSettings)

        query = super(SearchView, self).filter_query(query)
        if settings.respect_navroot:
            # If respect_navroot is enabled, return the filtered query
            # unchanged
            return query

        # Otherwise, if there wasn't a path filter in the query before,
        # remove the path filter that filter_query() put in.
        if query and 'path' in query and 'path' not in self.request.form:
            query.pop('path')
        return query

    def results(self, query=None, batch=True, b_size=10, b_start=0):
        """Get properly wrapped search results from the catalog.
        'query' should be a dictionary of catalog parameters.
        """
        self.solr_response = []
        # Disable theming for ajax requests
        if 'ajax' in self.request.form:
            del self.request.form['ajax']
            self.request.response.setHeader('X-Theme-Disabled', 'True')

        if query is None:
            query = {}

        query['b_start'] = b_start = int(b_start)
        query['b_size'] = b_size
        query = self.filter_query(query)

        if query is None:
            results = []
        else:
            query.update({'request_handler': 'hlsearch'})
            catalog = getToolByName(self.context, 'portal_catalog')
            try:
                results = catalog(**query)
            except ParseError:
                logger.exception('Exception while searching')
                return []
            except SolrConnectionException:
                logger.exception('Exception while searching')
                return []

        self.solr_response = results
        results = IContentListing(results)
        if batch:
            results = Batch(results, b_size, b_start)
        return results

    def snippets(self):
        """Return snippets with highlighted search terms in a dict with the
           item's UID as key and the snippet text as value.
        """
        if not safe_hasattr(self, 'solr_response'):
            return {}
        if safe_hasattr(self.solr_response, 'highlighting'):
            joined_snippets = {}
            for uid, snippets in self.solr_response.highlighting.items():
                joined_snippets[uid] = ' '.join([' '.join(snippet) for snippet
                                                 in snippets.values()]).strip()
            return joined_snippets
        return {}

    def suggestions(self):
        """Get suggestions from spellcheck component.
        """
        if not safe_hasattr(self, 'solr_response'):
            return []

        suggested_terms = []
        search_terms = [term.lower() for term in self.request.form.get(
            'SearchableText', '').split()]
        query_params = self.request.form.copy()

        if safe_hasattr(self.solr_response, 'spellcheck'):
            suggestions = self.solr_response.spellcheck.get('suggestions', [])
            for term in search_terms:
                if term in suggestions:
                    suggestion = suggestions[term]['suggestion']
                    query_params['SearchableText'] = suggestion[0]['word']
                    query_string = ''
                    for k, v in query_params.items():
                        if isinstance(v, list):
                            query_string += '&' + '&'.join(
                                ['%s=%s' % (k, vv) for vv in v])
                        else:
                            query_string += '&%s=%s' % (k, v)
                    suggested_terms.append((suggestion[0]['word'],
                                            query_string))
        return suggested_terms

    def breadcrumbs(self, item):
        if hasattr(item, 'is_external') and item.is_external():
            return None

        registry = getUtility(IRegistry)
        settings = registry.forInterface(ISearchSettings)
        maxb = settings.max_breadcrumbs

        if settings.path_based_breadcrumbs:
            portal_url = getToolByName(self.context, 'portal_url')
            portal = portal_url.getPortalObject()
            portal_path = portal_url.getPortalPath()
            path = item.getPath()[len(portal_path):]
            breadcrumbs = []
            url = portal.absolute_url()
            for part in path.split('/')[1:-1]:
                url = '%s/%s' % (url, part)
                breadcrumbs.append({
                    'Title': part,
                    'absolute_url': url,
                })
        else:
            obj = item.getObject()
            view = getMultiAdapter((obj, self.request),
                                   name='breadcrumbs_view')
            # cut off the item itself
            breadcrumbs = list(view.breadcrumbs())[:-1]

        if len(breadcrumbs) == 0:
            # don't show breadcrumbs if we only have a single element
            return None
        if len(breadcrumbs) > maxb:
            # if we have too long breadcrumbs, emit the middle elements
            empty = {'absolute_url': '', 'Title': unicode('…', 'utf-8')}
            breadcrumbs = [breadcrumbs[0], empty] + breadcrumbs[-maxb + 1:]
        return breadcrumbs

    def remove_path_link_info(self):
        query_dict = urlparse.parse_qs(self.request.get('QUERY_STRING', ''))

        if 'path' not in query_dict:
            return None

        path = query_dict['path']

        # Path may be a list, but wie deal only with the first item anyway
        path = isinstance(path, list) and path[0] or path
        item = self.context.unrestrictedTraverse(path)
        del query_dict['path']

        return {'search_url': '{0}?{1}'.format(self.request.get('ACTUAL_URL'),
                                               urllib.urlencode(query_dict, doseq=True)),
                'title': item.Title(),
                'url': item.absolute_url()}



from ftw.solr.testing import SOLR_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from unittest2 import TestCase
from plone.indexer.wrapper import IndexableObjectWrapper


class TestSnippetText(TestCase):

    layer = SOLR_FUNCTIONAL_TESTING

    def test_id_not_in_snippet_text(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        doc = portal[portal.invokeFactory('Document', 'doc1',
                                          title=u"Document 1")]
        wrapped = IndexableObjectWrapper(doc, portal.portal_catalog)
        self.assertNotIn('doc1', wrapped.snippetText)

    def test_title_not_in_snippet_text(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        doc = portal[portal.invokeFactory('Document', 'doc1',
                                          title=u"Document 1")]
        wrapped = IndexableObjectWrapper(doc, portal.portal_catalog)
        self.assertNotIn('Document 1', wrapped.snippetText)

    def test_searchwords_not_in_snippet_text(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        doc = portal[portal.invokeFactory('Document', 'doc1',
                                          title=u"Document 1",
                                          searchwords=u"Spam\nEggs")]
        wrapped = IndexableObjectWrapper(doc, portal.portal_catalog)
        self.assertNotIn('Spam', wrapped.snippetText)
        self.assertNotIn('Eggs', wrapped.snippetText)

    def test_body_text_in_snippet_text(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        doc = portal[portal.invokeFactory('Document', 'doc1',
                                          title=u"Document 1",
                                          text=u"Body Text")]
        wrapped = IndexableObjectWrapper(doc, portal.portal_catalog)
        self.assertIn('Body Text', wrapped.snippetText)

    def test_description_in_snippet_text(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        doc = portal[portal.invokeFactory('Document', 'doc1',
                                          title=u"Document 1",
                                          description=u"Description for "
                                                       "Document 1")]
        wrapped = IndexableObjectWrapper(doc, portal.portal_catalog)
        self.assertIn('Description for Document 1', wrapped.snippetText)

    def test_strip_html_tags(self):
        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        doc = portal[portal.invokeFactory('Document', 'doc1')]
        def SearchableText(self):
            return "<p>Body Text</p>"
        doc.SearchableText = SearchableText.__get__(doc, doc.__class__) 
        wrapped = IndexableObjectWrapper(doc, portal.portal_catalog)
        self.assertIn('Body Text', wrapped.snippetText)
        self.assertNotIn('<p>', wrapped.snippetText)
        self.assertNotIn('</p>', wrapped.snippetText)

    def test_searchable_text_indexer(self):
        from plone.indexer.interfaces import IIndexer
        from Products.ZCatalog.interfaces import IZCatalog
        from zope.interface import implements, Interface, alsoProvides
        from zope.component import provideAdapter, adapts
        class IMarker(Interface):
            pass
        class SearchableText(object):
            implements(IIndexer)
            adapts(IMarker, IZCatalog)
            def __init__(self, context, catalog):
                pass
            def __call__(self):
                return "Spam and eggs"
        provideAdapter(SearchableText, name="SearchableText")

        portal = self.layer['portal']
        setRoles(portal, TEST_USER_ID, ['Manager'])
        doc = portal[portal.invokeFactory('Document', 'doc1')]
        alsoProvides(doc, IMarker)
        wrapped = IndexableObjectWrapper(doc, portal.portal_catalog)
        self.assertIn('Spam and eggs', wrapped.snippetText)

    def test_non_archetypes_content(self):
        from Products.CMFCore.interfaces import IContentish
        from zope.interface import implements
        class DummyType(object):
            implements(IContentish)
            def SearchableText(self_):
                return "Dummy SearchableText"
            def Title(self_):
                return "Dummy"
            def Schema(self_):
                self.fail("Not an archetypes object.")
        portal = self.layer['portal']
        doc = DummyType()
        wrapped = IndexableObjectWrapper(doc, portal.portal_catalog)
        self.assertEquals(" SearchableText", wrapped.snippetText)



from ftw.solr.testing import SOLR_FUNCTIONAL_TESTING
from plone.app.testing import setRoles
from plone.app.testing import TEST_USER_ID
from unittest2 import TestCase
import transaction


class FunctionalTestCase(TestCase):
    layer = SOLR_FUNCTIONAL_TESTING

    def setUp(self):
        self.portal = self.layer['portal']

    def grant(self, *roles):
        setRoles(self.portal, TEST_USER_ID, list(roles))
        transaction.commit()



from AccessControl import getSecurityManager
from DateTime import DateTime
from zope.component import queryUtility
from collective.solr.interfaces import ISolrConnectionConfig
from collective.solr.mangler import iso8601date
from collective.solr.mangler import sort_aliases, query_args, ignored, ranges
from collective.solr.queryparser import quote
from collective.solr.utils import prepare_wildcard
from ftw.solr.patches.utils import isSimpleTerm
from ftw.solr.patches.utils import isSimpleSearch


def strip_wildcards(value):
    return value.replace('*', '').replace('?', '')


def strip_parens(value):
    return value.strip('()')


def searchterms_from_value(value):
    """Turn a search query into a list of search terms, removing
    parentheses, wildcards and quoting any special characters.
    """
    # remove any parens and wildcards, so quote() doesn't try to escape them
    value = strip_wildcards(strip_parens(value))
    # then quote the value
    value = quote(value)
    # and again strip parentheses that might have been added by quote()
    value = strip_parens(value)
    return value.split()


def leading_wildcards(value):
    """Prepend wildcards to each term for a string of search terms.
    (foo bar baz) -> (*foo *bar *baz)
    """
    search_terms = searchterms_from_value(value)
    value = ' '.join(['*%s' % term for term in search_terms])
    return "(%s)" % prepare_wildcard(value)


def trailing_wildcards(value):
    """Append wildcards to each term for a string of search terms.
    (foo bar baz) -> (foo* bar* baz*)
    """
    search_terms = searchterms_from_value(value)
    value = ' '.join(['%s*' % term for term in search_terms])
    return "(%s)" % prepare_wildcard(value)


def mangle_searchable_text_query(value, pattern):
    value = value.lower()

    value_lwc = leading_wildcards(value)
    value_twc = trailing_wildcards(value)
    value = strip_wildcards(value)

    value = pattern.format(
        value=quote(value),
        value_lwc=value_lwc,
        value_twc=value_twc)
    return value


def mangleQuery(keywords, config, schema):
    """ translate / mangle query parameters to replace zope specifics
        with equivalent constructs for solr """
    extras = {}
    for key, value in keywords.items():
        if key.endswith('_usage'):          # convert old-style parameters
            category, spec = value.split(':', 1)
            extras[key[:-6]] = {category: spec}
            del keywords[key]
        elif isinstance(value, dict):       # unify dict parameters
            keywords[key] = value['query']
            del value['query']
            extras[key] = value
        elif hasattr(value, 'query'):       # unify object parameters
            keywords[key] = value.query
            extra = dict()
            for arg in query_args:
                arg_val = getattr(value, arg, None)
                if arg_val is not None:
                    extra[arg] = arg_val
            extras[key] = extra
        elif key in ignored:
            del keywords[key]

    # find EPI indexes
    if schema:
        epi_indexes = {}
        for name in schema.keys():
            parts = name.split('_')
            if parts[-1] in ['string', 'depth', 'parents']:
                count = epi_indexes.get(parts[0], 0)
                epi_indexes[parts[0]] = count + 1
        epi_indexes = [k for k, v in epi_indexes.items() if v == 3]
    else:
        epi_indexes = ['path']

    for key, value in keywords.items():
        args = extras.get(key, {})
        if key == 'SearchableText':
            pattern = getattr(config, 'search_pattern', '')
            if pattern and isSimpleSearch(value):
                value = mangle_searchable_text_query(value, pattern)
                keywords[key] = set([value])    # add literal query parameter
                continue
            elif isSimpleTerm(value): # use prefix/wildcard search
                keywords[key] = '(%s* OR %s)' % (
                    prepare_wildcard(value), value)
                continue
        if key in epi_indexes:
            path = keywords['%s_parents' % key] = value
            del keywords[key]
            if 'depth' in args:
                depth = int(args['depth'])
                if depth >= 0:
                    if not isinstance(value, (list, tuple)):
                        path = [path]
                    tmpl = '(+%s_depth:[%d TO %d] AND +%s_parents:%s)'
                    params = keywords['%s_parents' % key] = set()
                    for p in path:
                        base = len(p.split('/'))
                        params.add(tmpl % (key, base, base + depth, key, p))
                del args['depth']
        elif key == 'effectiveRange':
            if isinstance(value, DateTime):
                steps = getattr(config, 'effective_steps', 1)
                if steps > 1:
                    value = DateTime(value.timeTime() // steps * steps)
                value = iso8601date(value)
            del keywords[key]
            keywords['effective'] = '[* TO %s]' % value
            keywords['expires'] = '[%s TO *]' % value
        elif key == 'show_inactive':
            del keywords[key]           # marker for `effectiveRange`
        elif 'range' in args:
            if not isinstance(value, (list, tuple)):
                value = [value]
            payload = map(iso8601date, value)
            keywords[key] = ranges[args['range']] % tuple(payload)
            del args['range']
        elif 'operator' in args:
            if isinstance(value, (list, tuple)) and len(value) > 1:
                sep = ' %s ' % args['operator'].upper()
                value = sep.join(map(str, map(iso8601date, value)))
                keywords[key] = '(%s)' % value
            del args['operator']
        elif key == 'allowedRolesAndUsers':
            if getattr(config, 'exclude_user', False):
                token = 'user$' + getSecurityManager().getUser().getId()
                if token in value:
                    value.remove(token)
        elif isinstance(value, DateTime):
            keywords[key] = iso8601date(value)
        elif not isinstance(value, basestring):
            assert not args, 'unsupported usage: %r' % args


# Allow more solr parameters
# e.g. 'qt'
def subtractQueryParameters(args, request_keywords=None):
    """ subtract parameters related to sorting and limiting search results
        from a given set of arguments, also removing them from the input """
    def get(name):
        for prefix in 'sort_', 'sort-':
            key = '%s%s' % (prefix, name)
            value = args.get(key, None)
            if value is not None:
                del args[key]
                return value
        return None
    params = {}
    index = get('on')
    if index:
        reverse = get('order') or ''
        reverse = reverse.lower() in ('reverse', 'descending')
        order = reverse and 'desc' or 'asc'
        params['sort'] = '%s %s' % (index, order)
    limit = get('limit')
    if limit:
        params['rows'] = int(limit)
    for key, value in args.items():
        if key in ('fq', 'fl', 'facet', 'qt'):
            params[key] = value
            del args[key]
        elif key.startswith('facet.') or key.startswith('facet_'):
            name = lambda facet: facet.split(':', 1)[0]
            if isinstance(value, list):
                value = map(name, value)
            elif isinstance(value, tuple):
                value = tuple(map(name, value))
            else:
                value = name(value)
            params[key.replace('_', '.', 1)] = value
            del args[key]
        elif key == 'b_start':
            params['start'] = int(value)
            del args[key]
        elif key == 'b_size':
            params['rows'] = int(value)
            del args[key]

    # Add default search handler if no search handler is specified in the 
    # query. With Solr 4 we have to disable the /select search handler to be
    # able to select another search handler with the 'qt' parameter. However
    # queries without a 'qt' parameter do no longer work.
    # TODO: In the future we should extend c.solr to handle multiple search
    # handlers by URL.
    if 'qt' not in params:
        params['qt'] = 'select'

    return params

# Remove facet.field parameters specifying an non-existing field.
def cleanupQueryParameters(args, schema):
    """ validate and possibly clean up the given query parameters using
        the given solr schema """
    sort = args.get('sort', None)
    if sort is not None:
        field, order = sort.split(' ', 1)
        if not field in schema:
            field = sort_aliases.get(field, None)
        fld = schema.get(field, None)
        if fld is not None and fld.indexed:
            args['sort'] = '%s %s' % (field, order)
        else:
            del args['sort']

    # Remove facet fields that are not in the solr schema
    facet_fields = args.get('facet.field', [])
    if not isinstance(facet_fields, list):
        facet_fields = [facet_fields]
    valid_facet_fields = []
    for facet_field in facet_fields:
        fld = schema.get(facet_field, None)
        if fld is not None and fld.indexed:
            valid_facet_fields.append(facet_field)
    if valid_facet_fields:
        args['facet.field'] = valid_facet_fields
    elif 'facet.field' in args:
        del args['facet.field']

    if 'facet.field' in args and not 'facet' in args:
        args['facet'] = 'true'

    return args


# Do not extend params['fq'] if it is a list as this would end
# in the request.
def optimizeQueryParameters(query, params):
    """ optimize query parameters by using filter queries for
        configured indexes """
    config = queryUtility(ISolrConnectionConfig)
    fq = []
    if config is not None:
        for idxs in config.filter_queries:
            idxs = set(idxs.split(' '))
            if idxs.issubset(query.keys()):
                fq.append(' '.join([query.pop(idx) for idx in idxs]))
    if 'fq' in params:
        if isinstance(params['fq'], list):
            params['fq'] = params['fq'] + fq
        else:
            params['fq'] = [params['fq']] + fq
    elif fq:
        params['fq'] = fq
    if not query:
        query['*'] = '*:*'      # catch all if no regular query is left...



from collective.solr import manager
from collective.solr.solr import SolrConnection


# TODO: Remove whole file in next major release.


class FtwSolrConnection(SolrConnection):
    pass


class SolrConnectionManager(manager.SolrConnectionManager):
    pass



from urllib import urlencode
from copy import deepcopy
from plone.memoize.view import memoize
from zope.component import getUtility, queryUtility
from zope.i18n import translate
from zope.i18nmessageid import Message

from collective.solr.browser import facets
from collective.solr.browser.facets import param, facetParameters
from collective.solr.interfaces import IFacetTitleVocabularyFactory
from collective.solr.interfaces import ISolrConnectionConfig


FACET_QUERY_POSITIONS = {
    '[NOW/DAY TO *]': 0,
    '[NOW/DAY-1DAY TO NOW/DAY]': 1,
    '[NOW/DAY-7DAYS TO *]': 2,
    '[NOW/DAY-1MONTH TO *]': 3,
    '[NOW/DAY-1YEAR TO *]': 4,
}


class SearchFacetsView(facets.SearchFacetsView):

    @memoize
    def facet_queries(self):
        """Extract facet queries from results.
        """
        results = self.kw.get('results', None)
        fcs = getattr(results, 'facet_counts', None)
        if results is None or fcs is None:
            return None

        fqueries = fcs.get('facet_queries', {})
        facets = {}
        for fquery, count in fqueries.items():
            if count < 1:
                continue
            facet_title, item_title = fquery.split(':', 1)
            if facet_title not in facets:
                facets[facet_title] = {}
            facets[facet_title][item_title] = count
        return facets

    def facets(self):
        """Prepare and return facetting info for the given SolrResponse """
        info = super(SearchFacetsView, self).facets()
        facet_queries = self.facet_queries()
        if not facet_queries:
            return info

        params = self.request.form.copy()
        if 'b_start' in params:
            del params['b_start'] # Clear the batch when limiting a result set
        fq = params.get('fq', [])
        if isinstance(fq, basestring):
            fq = params['fq'] = [fq]
        selected = set([facet.split(':', 1)[0] for facet in fq])

        for facet_title, facet_counts in facet_queries.items():
            if facet_title in selected:
                continue
            counts = []

            # Look up a vocabulary to provide a title for this facet
            vfactory = queryUtility(IFacetTitleVocabularyFactory, name=facet_title)
            if vfactory is None:
                # Use the default fallback
                vfactory = getUtility(IFacetTitleVocabularyFactory)
            vocabulary = vfactory(self.context)

            for name, count in facet_counts.items():
                p = deepcopy(params)
                p.setdefault('fq', []).append('%s:%s' % (facet_title, name.encode('utf-8')))

                title = name
                if name in vocabulary:
                    title = vocabulary.getTerm(name).title
                if isinstance(title, Message):
                    title = translate(title, context=self.request)

                counts.append(dict(
                    name=name,
                    title=title,
                    count=count,
                    query=urlencode(p, doseq=True),
                ))
            info.append(dict(title=facet_title, counts=sorted(counts,
                key=lambda x: FACET_QUERY_POSITIONS.get(x['name'], 100))))

        # Sort facets in the order stored in the configuration
        config = queryUtility(ISolrConnectionConfig)
        if config is not None:
            def pos(item):
                try:
                    return config.facets.index(item['title'])
                except ValueError:
                    return len(config.facets)
            info.sort(key=pos)

        return info

    def selected(self):
        """Determine selected facets and prepare links to clear them;
           this assumes that facets are selected using filter queries."""
        info = []
        facets = param(self, 'facet.field')
        facet_queries = self.facet_queries()
        fq = param(self, 'fq')
        for idx, query in enumerate(fq):
            field, value = query.split(':', 1)
            params = self.request.form.copy()
            params['fq'] = fq[:idx] + fq[idx+1:]

            # Facet queries are handled slightly different
            # (they are not enclosed by double quotes)
            if field in facet_queries:
                # Look up a vocabulary to provide a title for this facet
                vfactory = queryUtility(IFacetTitleVocabularyFactory, name=field)
                if vfactory is None:
                    # Use the default fallback
                    vfactory = getUtility(IFacetTitleVocabularyFactory)
                vocabulary = vfactory(self.context)
                if value in vocabulary:
                    value = vocabulary.getTerm(value).title
                if isinstance(value, Message):
                    value = translate(value, context=self.request)

                info.append(dict(title=field, value=value,
                    query=urlencode(params, doseq=True)))
                continue

            # Facet fields
            if field not in facets:
                params['facet.field'] = facets + [field]
            if value.startswith('"') and value.endswith('"'):
                # Look up a vocabulary to provide a title for this facet
                vfactory = queryUtility(IFacetTitleVocabularyFactory, name=field)
                if vfactory is None:
                    # Use the default fallback
                    vfactory = getUtility(IFacetTitleVocabularyFactory)
                vocabulary = vfactory(self.context)
                value = value[1:-1]
                if value in vocabulary:
                    value = vocabulary.getTerm(value.decode('utf8')).title
                if isinstance(value, Message):
                    value = translate(value, context=self.request)

                info.append(dict(title=field, value=value,
                    query=urlencode(params, doseq=True)))

        return info

    def facet_parameters(self):
        """Return the facet parameters to be queried for as an url-encoded
           string.
        """
        facets, dependencies = facetParameters(self)
        return urlencode({'facet': 'true', 'facet.field': facets, },
                         doseq=True)



from setuptools import setup, find_packages
import os

version = '1.7.2'

tests_require = [
    'ftw.builder',
    'ftw.testbrowser',
    'plone.app.dexterity',
    'plone.app.testing',
    ]

setup(name='ftw.solr',
      version=version,
      description="Solr integration for Plone using collective.solr",
      long_description=open("README.rst").read() + "\n" + \
          open(os.path.join("docs", "HISTORY.txt")).read(),

      classifiers=[
        "Framework :: Plone",
        "Programming Language :: Python",
        ],

      keywords='ftw solr',
      author='4teamwork AG',
      author_email='mailto:info@4teamwork.ch',
      url='https://github.com/4teamwork/ftw.solr',
      license='GPL2',

      packages=find_packages(exclude=['ez_setup']),
      namespace_packages=['ftw'],
      include_package_data=True,
      zip_safe=False,

      install_requires=[
        'collective.js.jqueryui',
        'collective.monkeypatcher',
        'collective.solr',
        'ftw.upgrade',
        'plone.api',
        'plone.app.contentlisting',
        'plone.app.search',
        'requests',
        'setuptools',
        'z3c.unconfigure',
        ],

      tests_require=tests_require,
      extras_require=dict(tests=tests_require),

      entry_points="""
      # -*- Entry points: -*-
      [z3c.autoinclude.plugin]
      target = plone
      """,
      )
